#include <signal.h>

#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <glade/glade.h>

#include <utillib.h>

#include "project.h"
#include "project-properties.h"
#include "gdownman.h"
#include "new-url.h"
#include "new-project.h"

#define ICON_SIZE 24

GType           gdownman_get_type           ();
static void     gdownman_class_init         (GdownmanClass *klass);
static void     gdownman_init               (Gdownman *gdownman);

static gint     delete_event_cb     (GtkWidget *window, GdkEventAny *e,
                                     gpointer data);

void    gdownman_quit               (GtkWidget *caller, gpointer data);
void    gdownman_select_all         (GtkWidget *caller, gpointer data);
void    gdownman_select_done        (GtkWidget *caller, gpointer data);
void    gdownman_select_stop        (GtkWidget *caller, gpointer data);
void    gdownman_select_queue       (GtkWidget *caller, gpointer data);
void    gdownman_select_active      (GtkWidget *caller, gpointer data);
void    gdownman_select_matched     (GtkWidget *caller, gpointer data);
static void     gdownman_checkmenuitem_config_notify
                                    (GConfClient *client,
                                     guint        cnxn_id,
                                     GConfEntry  *entry,
                                     gpointer     user_data);
void    gdownman_checkmenuitem_configure
                                    (GtkCheckMenuItem *checkmenuitem,
                                     GConfClient      *client,
                                     gchar            *config_key,
                                     GtkWidget        *widget,
                                     gboolean          default_value);
void    gdownman_checkmenuitem_switch   (GtkWidget *caller, gpointer data);
void    gdownman_servers_show           (GtkWidget *caller, gpointer data);
void    gdownman_project_change     (GtkWidget *caller, GtkNotebookPage *page,
                                     guint page_num,
                                     gpointer data);
void    gdownman_project_add        (Gdownman *gdownman, Project *project);
void    gdownman_project_remove     (Gdownman *gdownman, Project *project);
void    gdownman_project_new        (GtkWidget *caller, gpointer data);
void    gdownman_project_delete     (GtkWidget *caller, gpointer data);
void    gdownman_download_new       (GtkWidget *caller, gpointer data);
void    gdownman_project_properties (GtkWidget *caller, gpointer data);
void    gdownman_download_properties (GtkWidget *caller, gpointer data);
void    gdownman_download_start     (GtkWidget *caller, gpointer data);
void    gdownman_download_stop      (GtkWidget *caller, gpointer data);
void    gdownman_download_queue     (GtkWidget *caller, gpointer data);
void    gdownman_download_delete    (GtkWidget *caller, gpointer data);
void    gdownman_download_auto      (GtkWidget *caller, gpointer data);

Project *   gdownman_get_current_project    (Gdownman *gdownman);

void    gdownman_actions_update     (Gdownman *gdownman);
void    gdownman_disconnected       (Gdownman *gdownman);
void    gdownman_about              (GtkWidget *widget, gpointer data);

gboolean    gdownman_get_data               (GIOChannel *source,
                                             GIOCondition condition,
                                             gpointer data);

/* Menu and toolbar are stored here */
#include "gdownman_ui.h"

static gpointer parent_class = NULL;

GType gdownman_get_type ()
{
    static GType gdownman_type = 0;

    if (!gdownman_type)
    {
        static const GTypeInfo gdownman_info =
        {
            sizeof (GdownmanClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) gdownman_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Gdownman),
            0,              /* n_preallocs */
            (GInstanceInitFunc) gdownman_init
        };

        gdownman_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Gdownman",
                                               &gdownman_info,
                                               0);
    }

    return gdownman_type;
}

static void gdownman_class_init (GdownmanClass *klass)
{
    parent_class = g_type_class_peek_parent (klass);
}

static void gdownman_init (Gdownman *gdownman)
{
    gdownman->app = 0;
    gdownman->gnome_program = 0;
    gdownman->gnome_client = 0;
    gdownman->page_count = 0;
    gdownman->connected = FALSE;
    gdownman->fd = 0;
    gdownman->input_tag = 0;

    gdownman->projects = g_hash_table_new (NULL, NULL);
    gdownman->mimetype_cache = g_hash_table_new (g_str_hash, g_str_equal);
    gdownman->icon_theme = gnome_icon_theme_new ();

    gdownman->project_properties = NULL;
    gdownman->download_properties = 0;
    gdownman->servers = 0;
}

Gdownman *gdownman_new (int argc, char *argv[])
{
    Gdownman *gdownman = GDOWNMAN (g_object_new (gdownman_get_type (), NULL));

    if (signal (SIGPIPE, SIG_IGN) == SIG_ERR)
    {
        perror ("can't ignore SIGPIPE");
    }

    gnome_vfs_init();

    gdownman->gnome_program = gnome_program_init("gdownman", "0.0",
                                                 LIBGNOMEUI_MODULE,
                                                 argc, argv,
                                                 GNOME_PARAM_POPT_TABLE,
                                                 NULL, NULL);

    /* initialize glade for gnome */
    glade_gnome_init ();

    gdownman->app = gnome_app_new("gdownman", "Gnome Download Manager");

    gdownman->gconfclient = gconf_client_get_default ();
    gconf_client_add_dir (gdownman->gconfclient, "/apps/gdownman",
                          GCONF_CLIENT_PRELOAD_NONE, NULL);

	return gdownman;
}

void gdownman_delete(Gdownman *gdownman)
{
    if (gdownman->connected)
        close(gdownman->fd);
}

static gint
warning_dialog_cb (GtkWidget* window, GdkEventAny* e, gpointer data)
{
    return TRUE;
}

void
gdownman_ui_build (Gdownman *gdownman)
{
    GnomeApp *app = GNOME_APP(gdownman->app);
    GtkWidget *statusbar, *hbox;
    GtkCheckMenuItem *checkmenuitem;
    GConfClient *client = gdownman->gconfclient;
    gchar *config_key;

	g_signal_connect(G_OBJECT(app), "delete_event",
					G_CALLBACK(delete_event_cb), gdownman);

	gtk_window_set_default_size(GTK_WINDOW(app), 400, 350);
    gtk_window_set_icon_from_file (GTK_WINDOW (app), PIXMAPDIR "/gdownman.png",
                                   NULL);

	hbox = gtk_hbox_new(FALSE, FALSE);
	statusbar = gtk_statusbar_new();
	gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), TRUE);
	gtk_box_pack_start(GTK_BOX(hbox), statusbar, TRUE, TRUE, 0);
    gnome_app_set_statusbar_custom (app, hbox, statusbar);
    gtk_widget_show_all (hbox);
    gdownman->statusbar = hbox;

	gnome_app_create_menus_with_data(app, menu, gdownman);
	gnome_app_install_menu_hints(app, menu);

	gnome_app_create_toolbar_with_data(app, toolbar, gdownman);

	gdownman->toolbar = GTK_WIDGET(gnome_app_get_dock_item_by_name(
								GNOME_APP(app),	GNOME_APP_TOOLBAR_NAME));

    /* view->toolbar */
    checkmenuitem = GTK_CHECK_MENU_ITEM (view_menu[0].widget);
    config_key = "/apps/gdownman/show_toolbar";
    gdownman_checkmenuitem_configure (checkmenuitem,
                                      client,
                                      config_key,
                                      gdownman->toolbar,
                                      TRUE);

    /* view->statusbar */
    checkmenuitem = GTK_CHECK_MENU_ITEM (view_menu[1].widget);
    config_key = "/apps/gdownman/show_statusbar";
    gdownman_checkmenuitem_configure (checkmenuitem,
                                      client,
                                      config_key,
                                      gdownman->statusbar,
                                      TRUE);

	gdownman->notebook = gtk_notebook_new();
	gnome_app_set_contents(app, gdownman->notebook);

    g_signal_connect(G_OBJECT(gdownman->notebook), "switch_page",
                     G_CALLBACK(gdownman_project_change), gdownman);

//FIXME: FIX GNOME-VFS!
    {
        GtkWidget *dialog;

        dialog = gtk_message_dialog_new (GTK_WINDOW (app),
                                         GTK_DIALOG_DESTROY_WITH_PARENT
                                         | GTK_DIALOG_NO_SEPARATOR,
                                         GTK_MESSAGE_WARNING,
                                         GTK_BUTTONS_NONE,
"This version of Gnome Download Manager DOES NOT SUPPORT RESUME, if you STOP a \
transfer, you will need to start it from the begining!\n\n This version is for \
help testing the gnome-vfs backends, so report any problems. Thanks!");

        gtk_window_set_transient_for (GTK_WINDOW (dialog),
                                      GTK_WINDOW (app));
        g_signal_connect (G_OBJECT (dialog), "delete_event",
                          G_CALLBACK (warning_dialog_cb),
                          NULL);
        gtk_widget_show (dialog);
    }

//	gnome_app_enable_layout_config(app, TRUE);
}

void gdownman_sm_init(Gdownman *gdownman)
{
	gdownman->gnome_client = gnome_master_client();
}

void
gdownman_ui_show (Gdownman *gdownman)
{
    /* Workaround silly GnomeApp implementation */
    gboolean visible;

    visible = GTK_WIDGET_VISIBLE (gdownman->toolbar);
    gtk_widget_show (gdownman->app);
    if (!visible)
        gtk_widget_hide (gdownman->toolbar);
}

void gdownman_quit(GtkWidget * caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;

//FIXME: save settings
	gdownman_delete(gdownman);
	gtk_main_quit();
}

static gint
delete_event_cb(GtkWidget* window, GdkEventAny* e, gpointer data)
{
	gdownman_quit(window, data);

	// Do not destroy it, we did at gdownman_quit()
	return TRUE;
}

void
gdownman_about (GtkWidget *widget, gpointer data)
{
	Gdownman *gdownman = GDOWNMAN (data);
    static GtkWidget *about = NULL;
    GdkPixbuf *logo;
    const char *authors[] = {
            "Manuel Clos <llanero@jazzfree.com>", NULL
        };

    if (about != NULL)
    {
        gtk_window_present (GTK_WINDOW (about));
        return;
    }

    logo = gdk_pixbuf_new_from_file (PIXMAPDIR "/gdownman.png", NULL);

    about = gnome_about_new ("Download Manager",
                             VERSION,
                             "Copyright (C) 2001 Manuel Clos",
                             _("A Download Manager for GNOME."),
                             authors,
                             NULL, /* documenters */
                             NULL, /* translator_credits */
                             logo);

    if (logo != NULL)
        g_object_unref (logo);

    gtk_window_set_destroy_with_parent (GTK_WINDOW (about), TRUE);
    gtk_window_set_transient_for (GTK_WINDOW (about),
                                  GTK_WINDOW (gdownman->app));

    g_signal_connect (G_OBJECT (about), "destroy",
                      G_CALLBACK (gtk_widget_destroyed),
                      &about);

    gtk_widget_show_all (about);
}

void
gdownman_checkmenuitem_configure (GtkCheckMenuItem *checkmenuitem,
                                  GConfClient      *client,
                                  gchar            *config_key,
                                  GtkWidget        *widget,
                                  gboolean          default_value)
{
    GConfValue* value = NULL;
    gboolean visible;
    guint notify_id;

    g_object_set_data (G_OBJECT (checkmenuitem), "updating",
                       GINT_TO_POINTER (1));

    value = gconf_client_get (client, config_key, NULL);
    if (value)
    {
        visible = gconf_client_get_bool (client, config_key, NULL);
        gtk_check_menu_item_set_active (checkmenuitem, visible);

        if (visible)
            gtk_widget_show (widget);
        else
            gtk_widget_hide (widget);
    }
    else
    {
        gtk_widget_show (widget);
        gtk_check_menu_item_set_active (checkmenuitem, default_value);
    }

    notify_id = gconf_client_notify_add (client,
                                         config_key,
                                         gdownman_checkmenuitem_config_notify,
                                         checkmenuitem,
                                         NULL, NULL);

    g_object_set_data (G_OBJECT (checkmenuitem), "notify_id",
                       GUINT_TO_POINTER (notify_id));
    g_object_set_data (G_OBJECT (checkmenuitem), "client", client);
    g_object_set_data (G_OBJECT (checkmenuitem), "config_key", config_key);
    g_object_set_data (G_OBJECT (checkmenuitem), "widget", widget);
    g_object_set_data (G_OBJECT (checkmenuitem), "updating",
                       GINT_TO_POINTER (0));
}

static void
gdownman_checkmenuitem_config_notify (GConfClient *client,
                                      guint        cnxn_id,
                                      GConfEntry  *entry,
                                      gpointer     user_data)
{
    GtkCheckMenuItem *check = user_data;

    g_return_if_fail (GTK_IS_CHECK_MENU_ITEM (check));

    if (GPOINTER_TO_INT (g_object_get_data (G_OBJECT (check), "updating")))
    {
        g_object_set_data (G_OBJECT (check), "updating", GINT_TO_POINTER (0));
        return;
    }

    g_object_set_data (G_OBJECT (check), "updating", GINT_TO_POINTER (1));

    if (gconf_entry_get_value (entry) == NULL)
    {
        return;
    }
    else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_BOOL)
    {
        GtkWidget *widget;
        gboolean visible;

        visible = gconf_value_get_bool (gconf_entry_get_value (entry));
        gtk_check_menu_item_set_active (check, visible);
        widget = g_object_get_data (G_OBJECT (check), "widget");
        if (visible)
            gtk_widget_show (widget);
        else
            gtk_widget_hide (widget);
    }
g_print ("notify end\n");
}

void
gdownman_checkmenuitem_switch (GtkWidget *caller, gpointer data)
{
    GtkCheckMenuItem *check = GTK_CHECK_MENU_ITEM (caller);
    GConfClient *client;
    GtkWidget *widget;
    gboolean visible;
    gchar *config_key;

    if (GPOINTER_TO_INT (g_object_get_data (G_OBJECT (check), "updating")))
    {
        g_object_set_data (G_OBJECT (check), "updating", GINT_TO_POINTER (0));
        return;
    }

    widget = g_object_get_data (G_OBJECT (check), "widget");
    visible = gtk_check_menu_item_get_active (check);
    if (visible)
        gtk_widget_show (widget);
    else
        gtk_widget_hide (widget);

    client = g_object_get_data (G_OBJECT (check), "client");
    config_key = g_object_get_data (G_OBJECT (check), "config_key");
    g_object_set_data (G_OBJECT (check), "updating", GINT_TO_POINTER (1));
    gconf_client_set_bool (client, config_key, visible, NULL);
}

void
gdownman_servers_show (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = GDOWNMAN (data);

    if (!gdownman->servers)
    {
        gdownman->servers = servers_new (gdownman->fd,
                                         GTK_WINDOW(gdownman->app));
    }

    servers_start (gdownman->servers);
}

void gdownman_project_change(GtkWidget *caller, GtkNotebookPage *page,
                             guint page_num, gpointer data)
{
    Gdownman *gdownman = (Gdownman *)data;
    Project *project;
	GtkWidget *child;

	child = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gdownman->notebook),
                                      page_num);

    if (child == NULL)
        project = NULL;
    else
        project = (Project *) g_object_get_data(G_OBJECT (child), "project");

    if (project == NULL)
        return;

    if (gdownman->project_properties != NULL)
    {
        project_properties_project_set (
                PROJECT_PROPERTIES (gdownman->project_properties),
                project);
    }

    if (gdownman->download_properties != NULL)
    {
        project_download_properties (project, gdownman->download_properties);
    }
}

void gdownman_project_add(Gdownman *gdownman, Project *project)
{
	GtkWidget *tab_label = gtk_label_new(project_name(project));
	GtkWidget *child = project_view_new(project);

    g_hash_table_insert (gdownman->projects,
                         GINT_TO_POINTER (project_id (project)),
                         project);

	g_object_set_data(G_OBJECT(child), "project", project);

	gtk_notebook_append_page(GTK_NOTEBOOK(gdownman->notebook),
							child, tab_label);

	gdownman->page_count++;

	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(gdownman->notebook),
						(gdownman->page_count < 2) ? FALSE : TRUE);

	gtk_widget_show(child);
}

void gdownman_project_remove(Gdownman *gdownman, Project *project)
{
	GtkWidget *child;
    gboolean removed;
	int i;

    removed = g_hash_table_remove (gdownman->projects,
                                   GINT_TO_POINTER (project_id (project)));

    if (!removed)
        g_warning ("The project was not found in the hash table while trying to remove it.");

	for (i = 0; i < gdownman->page_count; i++)
	{
		child = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gdownman->notebook), i);
		if (project == g_object_get_data(G_OBJECT(child), "project"))
		{
			gtk_notebook_remove_page(GTK_NOTEBOOK(gdownman->notebook), i);
            gdownman->page_count--;
            gtk_notebook_set_show_tabs (GTK_NOTEBOOK(gdownman->notebook),
                                    (gdownman->page_count < 2) ? FALSE : TRUE);
			break;
		}
	}

	project_delete(project);
}

void gdownman_connect_local(Gdownman *gdownman)
{
    GError *error = NULL;
    gboolean connected = FALSE;
    gboolean exec_error = FALSE;
    guint tries = 5;

    gdownman->fd = dm_client_local_connect("gdownman");
    if (gdownman->fd >= 0)
    {
        connected = TRUE;
    }

    while ((tries > 0) && (!connected))
    {
        tries--;

        if (!dm_client_daemon_start (&error))
        {
            exec_error = TRUE;
            break;
        }

        usleep (500000);

        gdownman->fd = dm_client_local_connect ("gdownman");
        if (gdownman->fd >= 0)
        {
            connected = TRUE;
            break;
        }
    }

	if (gdownman->fd < 0)
	{
        GtkWidget *dialog;

        dialog = gtk_message_dialog_new (GTK_WINDOW (gdownman->app),
                            GTK_DIALOG_MODAL,
                            GTK_MESSAGE_ERROR,
                            GTK_BUTTONS_OK,
                            "%s%s%s",
                            "Can't connect to the download daemon",
                            (exec_error) ? " and autostart failed:\n\n" : ".",
                            (exec_error) ? error->message : "");

		gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);

        gdownman_actions_update (gdownman);
		return;
	}

	gdownman->connected = TRUE;

    gdownman->local_channel = g_io_channel_unix_new (gdownman->fd);
    gdownman->input_tag = g_io_add_watch_full (gdownman->local_channel,
                                               G_PRIORITY_LOW,
                                               G_IO_IN,
                                               gdownman_get_data,
                                               gdownman,
                                               NULL);
	dm_command_events(gdownman->fd, 1, NULL);
	dm_command_batchprogress(gdownman->fd, 1, NULL);
	dm_command_batchstatus(gdownman->fd, 1, NULL);
}

void
gdownman_project_new (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = GDOWNMAN (data);

    NewProject *newproject = new_project_new (gdownman->fd);

    if (newproject)
        new_project_start (newproject);
}

void gdownman_project_delete(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;
	Project *project = gdownman_get_current_project(gdownman);

	dm_command_project_delete(gdownman->fd, project_id(project));
}

void gdownman_download_new(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;
    Project *project = gdownman_get_current_project (gdownman);
    NewUrl *newurl;

    newurl = new_url_new (project_id (project), project->settings,
                          gdownman->fd);

    if (newurl)
        new_url_start (newurl);
}

void
gdownman_download_properties (GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = GDOWNMAN (data);
    Project *project = gdownman_get_current_project (gdownman);

    if (gdownman->download_properties == NULL)
    {
        gdownman->download_properties = download_properties_new (
                                            gdownman->fd,
                                            GTK_WINDOW(gdownman->app));
    }

    download_properties_start (gdownman->download_properties);

    if (project)
    {
        project_download_properties (project, gdownman->download_properties);
    }
}

void gdownman_project_properties(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *)data;

    if (gdownman->project_properties == NULL)
    {
        gdownman->project_properties = project_properties_new(
                                            gdownman_get_current_project(gdownman),
                                            "Project properties",
                                            GTK_WINDOW(gdownman->app));
    }

    gtk_window_present (GTK_WINDOW (gdownman->project_properties));
}

void gdownman_send(Gdownman *gdownman, char *data)
{
	write(gdownman->fd, data, strlen(data) + 1);
}

Project * gdownman_get_current_project(Gdownman *gdownman)
{
    Project *project;
	int i = gtk_notebook_get_current_page(GTK_NOTEBOOK(gdownman->notebook));
	GtkWidget *child;
	
	child = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gdownman->notebook), i);

    if (child == NULL)
        project = NULL;
    else
        project = (Project *)g_object_get_data(G_OBJECT(child), "project");

    return project;
}

Project *
gdownman_project_from_id (Gdownman *gdownman, guint id)
{
    Project *project;

    project = g_hash_table_lookup (gdownman->projects,
                                   GINT_TO_POINTER (id));

    return project;
}

void gdownman_select_all(GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_all (gdownman_get_current_project(gdownman));
}

void gdownman_select_done (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_status (gdownman_get_current_project(gdownman),
                           STATUS_DONE);
}

void gdownman_select_stop (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_status (gdownman_get_current_project(gdownman),
                           STATUS_STOPPED);
}

void gdownman_select_queue (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_status (gdownman_get_current_project(gdownman),
                           STATUS_QUEUED);
}

void
gdownman_select_active (GtkWidget *caller, gpointer data)
{
    Gdownman *gdownman = (Gdownman *) data;

    project_select_active (gdownman_get_current_project(gdownman));
}

void
gdownman_select_matched (GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = GDOWNMAN (data);
	GtkWidget *dialog, *label, *entry, *hint_pixmap, *hint_label, *hbox;
	const gchar *text;

	dialog = gtk_dialog_new_with_buttons ("Select downloads",
					GTK_WINDOW(gdownman->app), 0,
					GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
					GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);

	label = gtk_label_new ("<big><b>File is:</b></big>");
	gtk_label_set_use_markup (GTK_LABEL(label), TRUE);
	entry = gtk_entry_new ();

    hint_pixmap = gtk_image_new_from_stock (GTK_STOCK_DIALOG_INFO,
                                            GTK_ICON_SIZE_SMALL_TOOLBAR);
    hint_label = gtk_label_new ("*.bz2 will match all bz2 files");
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);
    gtk_container_set_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);

    gtk_box_pack_start (GTK_BOX (hbox), hint_pixmap, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (hbox), hint_label, FALSE, FALSE, 0);

	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), entry);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), hbox);
	gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
	gtk_widget_grab_focus(entry);

	gtk_widget_show_all (dialog);

	switch(gtk_dialog_run(GTK_DIALOG(dialog)))
	{
		case GTK_RESPONSE_ACCEPT:
			text = gtk_entry_get_text(GTK_ENTRY(entry));
			if (strlen (text) > 0)
                project_select_matched (gdownman_get_current_project(gdownman),
                                        text);
			break;

		default:
			break;
	}

	gtk_widget_destroy(dialog);
}


void gdownman_download_start(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_START, gdownman->fd);
}

void gdownman_download_stop(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_STOP, gdownman->fd);
}

void gdownman_download_queue(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_QUEUE, gdownman->fd);
}

void gdownman_download_delete(GtkWidget *caller, gpointer data)
{
	Gdownman *gdownman = (Gdownman *) data;

	project_download(gdownman_get_current_project(gdownman),
					DOWNLOAD_DELETE, gdownman->fd);
}

void
gdownman_disconnected (Gdownman *gdownman)
{
    gdownman->connected = FALSE;
    gdownman_actions_update (gdownman);
}

void
gdownman_actions_update (Gdownman *gdownman)
{
    gboolean state = gdownman->connected;
    gint i;

    /* Disable toolbar */
    gtk_widget_set_sensitive (GTK_WIDGET (gdownman->toolbar), state);

    /* Disable _some_ file menus */
    gtk_widget_set_sensitive (GTK_WIDGET (file_menu[0].widget), state);
    gtk_widget_set_sensitive (GTK_WIDGET (file_menu[1].widget), state);

    /* Disable edit menus */
    for (i = 0; i < 6; i++)
    {
        gtk_widget_set_sensitive (GTK_WIDGET (edit_menu[i].widget), state);
    }

    /* Disable project menus */
    for (i = 0; i < 2; i++)
    {
        gtk_widget_set_sensitive (GTK_WIDGET (project_menu[i].widget), state);
    }
    /* Disable download menus */
    for (i = 0; i < 5; i++)
    {
        gtk_widget_set_sensitive (GTK_WIDGET (download_menu[i].widget), state);
    }

    /* Disable projects */
    gtk_widget_set_sensitive (gdownman->notebook, state);

    /* Disable download-properties dialog */
    project_properties_sensitive_set (
            PROJECT_PROPERTIES (gdownman->project_properties), state);

    /* Disable download-properties dialog */
    download_properties_sensitive_set (gdownman->download_properties, state);
}

GdkPixbuf *
gdownman_pixbuf_cache_get_mime_type (Gdownman *gdownman, gchar *filename)
{
    GHashTable *hash = gdownman->mimetype_cache;
    const gchar *mime_type;
    GdkPixbuf *pixbuf;
    GdkPixbuf *scaled;
    gchar *icon_name;
    gchar *icon_path;
    gint w, h;

    mime_type = gnome_vfs_get_mime_type (filename);

    if (!mime_type)
        return NULL;

    pixbuf = g_hash_table_lookup (hash, mime_type);

    if (pixbuf)
    {
        g_object_ref (pixbuf);
        return pixbuf;
    }

    /* Oh :( we wasted time looking at the hash table ... */
    icon_name = gnome_icon_lookup (gdownman->icon_theme, NULL, NULL, NULL, NULL,
                                   mime_type,
                                   GNOME_ICON_LOOKUP_FLAGS_NONE,
                                   NULL);

    if (!icon_name)
        return NULL;

    icon_path = gnome_icon_theme_lookup_icon (gdownman->icon_theme, icon_name,
                                              ICON_SIZE,
                                              NULL,
                                              NULL);

    if (!icon_path)
        return NULL;

    /* Wow! we are halfway to get the icon! */
    pixbuf = gdk_pixbuf_new_from_file (icon_path, NULL);

    if (!pixbuf)
        return NULL;

    gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &w, &h);
    scaled = gdk_pixbuf_scale_simple (pixbuf, w, h, GDK_INTERP_BILINEAR);
    g_object_unref (pixbuf);

    if (!scaled)
        return NULL;

    /* lots of work! let's save it for next time */
    g_hash_table_insert (hash, g_strdup (mime_type), scaled);
    g_object_ref (scaled);

    return scaled;
}

gboolean
gdownman_get_data (GIOChannel *source, GIOCondition condition, gpointer data)
{
    Gdownman *gdownman = GDOWNMAN (data);
	Project *project;
	unsigned int project_id;
	GtkWidget *dialog;
	char s[2048];
    char localpath[2048];
    int max, current;
	long int i,j;
	double k,l;
	double percent;
	int id, status;
    guint maxdownloads;
    GIOStatus iostatus;
    GString *buffer = g_string_new (NULL);
    gint terminator_pos;
    GError *error = NULL;
    ProjectItem *item;
    gchar *file;
    
//FIXME: utillib func !!!!!!!!!!!!!!!!!!!

//FIXME: Try to restart in 30 seconds
    if ((condition & G_IO_ERR) || (condition & G_IO_PRI) ||
        (condition & G_IO_NVAL) || (condition & G_IO_OUT))
    {
        g_print ("downmand input: condition = %u\n", condition);
        gdownman_disconnected (gdownman);
        return FALSE;
    }
    else if (condition & G_IO_IN)
    {
        iostatus = g_io_channel_read_line_string (source, buffer,
                                                  &terminator_pos,
                                                  &error);

        if (iostatus == G_IO_STATUS_EOF)
        {
            g_printerr ("Connection to downmand lost...\n");
            gdownman_disconnected (gdownman);
            return FALSE;
        }
        else if ((iostatus == G_IO_STATUS_AGAIN) || (iostatus = G_IO_STATUS_ERROR))
        {
            g_string_free (buffer, TRUE);
            return TRUE;
        }
        else
        {
            /* Go on with parsing the data */
        }
    }
    else if (condition & G_IO_HUP)
    {
        g_print ("downmand input: condition = %u\n", condition);
        return FALSE;
    }

//    g_print ("GOT: %s\n", buffer->str);

    if (sscanf (buffer->str, "%li", &i) == 0)
	    return TRUE;

    switch(i)
    {
        case DM_COMMAND_ITEM_NEW:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %s %i %s %lf %lf %u",
                        &project_id,
                        &item->id,
                        localpath,
                        &item->status,
                        s,
                        &item->downloaded,
                        &item->total,
                        &item->speedlimit) < 7)
            {
                g_free (item);
                return TRUE;
            }
            item->localpath = g_strdup (localpath);
            item->url = g_strdup (s);

            if (item->url[strlen (item->url) - 1] == '/')
                file = "index.html";
            else
                file = (strrchr(item->url, '/')) ? strrchr(item->url, '/') + 1 : item->url;

            g_string_printf (buffer, "%s/%s", item->localpath, file);
            item->mimetype_pixbuf = gdownman_pixbuf_cache_get_mime_type (
                                                    gdownman,
                                                    buffer->str);

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_NEW;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
            break;

        case DM_COMMAND_BATCH_PROGRESS:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %i %i %lf %lf",
                        &project_id,
                        &item->id,
                        &item->status,
                        &item->speed,
                        &item->downloaded,
                        &item->total) < 6)
            {
                g_free (item);
                return TRUE;
            }

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_UPDATE;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
            break;

        case DM_COMMAND_ITEM_DONE:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %i %lf %lf",
                        &project_id,
                        &item->id,
                        &item->status,
                        &item->downloaded,
                        &item->total) < 5)
            {
                g_free (item);
                return TRUE;
            }

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_DONE;
            item->speed = -1;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
			break;

		case DM_COMMAND_ITEM_STOP:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i %i %lf %lf",
                        &project_id,
                        &item->id,
                        &item->status,
                        &item->downloaded,
                        &item->total) < 5)
            {
                g_free (item);
                return TRUE;
            }

            if (item->total > 0)
                item->percent = (item->downloaded/item->total)*100;
            else
                item->percent = 0;

            item->type = PROJECT_ITEM_STOP;
            item->speed = -1;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
			break;

		case DM_COMMAND_ITEM_DELETE:
            item = g_malloc (sizeof (ProjectItem));

            if (sscanf (buffer->str + 4, "%u %i",
                        &project_id,
                        &item->id) < 2)
            {
                g_free (item);
                return TRUE;
            }

            item->type = PROJECT_ITEM_DELETE;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
			break;

        case DM_COMMAND_ITEM_SPEED:
            item = g_new0 (ProjectItem, 1);

            if (sscanf (buffer->str + 4, "%u %i %i",
                        &project_id,
                        &item->id,
                        &item->speedlimit) < 3)
            {
                g_free (item);
                return TRUE;
            }

            item->type = PROJECT_ITEM_SPEEDLIMIT;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
                project_queue_item (project, item);
            break;

		case DM_COMMAND_BATCH_STATUS:
//FIXME: utillib func for parsing
		    if (sscanf (buffer->str + 4, "%u %li %li %lf %lf",
						&project_id, &i, &j, &k, &l) < 5)
				return TRUE;

			if (l > 0)
				percent = (k/l)*100;
			else
				percent = 0;

			project = gdownman_project_from_id(gdownman, project_id);

			if (project)
			{
                g_string_printf (buffer,
                                 "Downloading %li of %li files - %.1f%% done",
                                 i,
                                 i + j,
                                 percent);
                project_status_set (project, buffer->str);
			}
			break;

        case DM_COMMAND_SERVER_NEW:
            if (!gdownman->servers)
                return TRUE;

		    if (sscanf (buffer->str + 4, "%s %i %u", s, &max, &current) < 3)
				return TRUE;

            servers_add (gdownman->servers, s, max, current);
            break;

        case DM_COMMAND_SERVER_MAX_CONNECTIONS:
            if (!gdownman->servers)
                return TRUE;

		    if (sscanf (buffer->str + 4, "%s %i", s, &max) < 2)
				return TRUE;

            servers_max_connections (gdownman->servers, s, max);
            break;

        case DM_COMMAND_SERVER_CURRENT_CONNECTIONS:
            if (!gdownman->servers)
                return TRUE;

		    if (sscanf (buffer->str + 4, "%s %u", s, &current) < 2)
				return TRUE;

            servers_current_connections (gdownman->servers, s, current);
            break;

		case DM_COMMAND_PROJECT_NEW:
		    if (sscanf (buffer->str + 4, "%i %s", &id, s) < 2)
				return TRUE;

			gdownman_project_add (gdownman,
                      project_new (id, s, gdownman->fd, gdownman->gconfclient));

			break;

		case DM_COMMAND_PROJECT_DELETE:
		    if (sscanf (buffer->str + 4, "%i", &id) < 1)
				return TRUE;

			project = gdownman_project_from_id(gdownman, id);

			if (project)
				gdownman_project_remove(gdownman, project);
			else
				g_print("Delete project error: %i\n", id);
			break;

        case DM_COMMAND_PROJECT_PATH:
            if (sscanf (buffer->str + 4, "%i %s", &project_id, s) < 2)
                return TRUE;

            project = gdownman_project_from_id (gdownman, project_id);

            if (project)
            {
                project_auto_set (project, status);
                project_path_set (project, s);
            }
            break;

		case DM_COMMAND_PROJECT_AUTO:
		    if (sscanf (buffer->str + 4, "%i %u", &project_id, &status) < 2)
				return TRUE;

			project = gdownman_project_from_id (gdownman, project_id);

			if (project)
            {
                project_auto_set (project, status);
            }
			break;

        case DM_COMMAND_PROJECT_MAXTRANSFERS:
		    if (sscanf (buffer->str + 4, "%i %u", &project_id, &maxdownloads) < 2)
				return TRUE;

			project = gdownman_project_from_id (gdownman, project_id);

			if (project)
            {
                project_maxdownloads_set (project, maxdownloads);
            }
			break;

		case DM_COMMAND_ERROR:
            dialog = gtk_message_dialog_new (GTK_WINDOW(gdownman->app),
                            GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
                            GTK_MESSAGE_ERROR,
                            GTK_BUTTONS_CLOSE,
                            "%s", buffer->str + 4);

			g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
								G_CALLBACK (gtk_widget_destroy),
								GTK_OBJECT (dialog));

			gtk_widget_show(dialog);
			break;

	    default:
	        g_print ("Not handled response!!!\n");
            g_print ("GOT: %s\n", buffer->str);
	        break;
	}
    return TRUE;
}
