#include <gnome.h>
#include <gconf/gconf.h>
#include <libgnomevfs/gnome-vfs.h>

#include <utillib.h>

#include "gdownman-marshal.h"
//#include "gtkcellrendererprogressbar.h"

#include "project-settings.h"
#include "project.h"


GType           project_get_type           ();
static void     project_class_init         (ProjectClass *klass);
static void     project_init               (Project *project);

void        status_label_size_request   (GtkWidget *widget,
                                         GtkRequisition  *requisition,
                                         gpointer data);
void        project_selection_changed   (GtkTreeSelection *selection,
                                         gpointer data);
static void project_view_add_columns    (Project *project);
void        project_column_format_progress (GtkTreeViewColumn *tree_column,
                                         GtkCellRenderer *cell,
                                         GtkTreeModel *tree_model,
                                         GtkTreeIter *iter,
                                         gpointer data);
void        project_column_format_value (GtkTreeViewColumn *tree_column,
                                         GtkCellRenderer *cell,
                                         GtkTreeModel *tree_model,
                                         GtkTreeIter *iter,
                                         gpointer data);
void        project_column_format_status    (GtkTreeViewColumn *tree_column,
                                             GtkCellRenderer *cell,
                                             GtkTreeModel *tree_model,
                                             GtkTreeIter *iter,
                                             gpointer data);
void        project_column_format_speed (GtkTreeViewColumn *tree_column,
                                         GtkCellRenderer *cell,
                                         GtkTreeModel *tree_model,
                                         GtkTreeIter *iter,
                                         gpointer data);
void        project_download_auto       (GtkWidget *caller, gpointer data);
void        project_download_auto_set   (Project *project);
gboolean    project_process_updates     (gpointer data);
void        project_item_new            (Project *project,
                                         ProjectItem *item);
void        project_item_update         (Project *project,
                                         ProjectItem *item);
void        project_item_speedlimit     (Project *project, ProjectItem *item);
void        project_download_properties (Project *project,
                                     DownloadProperties *download_properties);

enum
{
	COLUMN_ID,
	COLUMN_MIMETYPE_ICON,
	COLUMN_FILE,            /* filename only */
	COLUMN_FILENAME,        /* complete path to file */
	COLUMN_STATUS_ICON,
	COLUMN_STATUS,
	COLUMN_PERCENT,
	COLUMN_SPEED,
	COLUMN_DOWNLOADED,
	COLUMN_TOTAL,
	COLUMN_URL,
	COLUMN_SPEEDLIMIT,
	NUM_COLUMNS
};

enum {
    PROJECT_PATH,
    PROJECT_AUTODOWNLOAD,
    PROJECT_MAXDOWNLOADS,
    LAST_SIGNAL
};

static gint project_signals[LAST_SIGNAL] = { 0 };

static gpointer parent_class = NULL;

GType project_get_type ()
{
    static GType project_type = 0;

    if (!project_type)
    {
        static const GTypeInfo project_info =
        {
            sizeof (ProjectClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) project_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Project),
            0,              /* n_preallocs */
            (GInstanceInitFunc) project_init
        };

        project_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Project",
                                               &project_info,
                                               0);
    }

    return project_type;
}

static void project_class_init (ProjectClass *klass)
{
    parent_class = g_type_class_peek_parent (klass);

    project_signals[PROJECT_AUTODOWNLOAD] = g_signal_new ("project_autodownload",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, project_autodownload_signal),
            NULL, NULL,
            gdownman_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    project_signals[PROJECT_MAXDOWNLOADS] = g_signal_new ("project_maxdownloads",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, project_maxdownloads_signal),
            NULL, NULL,
            gdownman_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    klass->project_autodownload_signal = NULL;
    klass->project_maxdownloads_signal = NULL;
}

static void project_init (Project *project)
{
	project->hash = g_hash_table_new (NULL, NULL);
    project->updates_waiting = 0;
    project->updates_queue = g_queue_new ();
	project->action = -1;
	project->auto_button = 0;
	project->auto_label = 0;
	project->auto_yes = 0;
	project->auto_no = 0;
	project->pixbuf_mimetype = 0;
	project->pixbuf_exec = 0;
	project->pixbuf_stop = 0;
	project->pixbuf_queue = 0;
	project->pixbuf_warning = 0;
	project->pixbuf_error = 0;
	project->pixbuf_done = 0;
    project->settings = project_settings_new ();
    project->selection = 0;
    project->download_properties = 0;
}

Project *
project_new (guint id, gchar *name, guint fd, GConfClient *client)
{
    Project *project = PROJECT (g_object_new (project_get_type (), NULL));

    /* create tree model */
    project->model = GTK_TREE_MODEL (gtk_list_store_new (NUM_COLUMNS,
                                                         G_TYPE_UINT,
                                                         GDK_TYPE_PIXBUF,
                                                         G_TYPE_STRING,
                                                         G_TYPE_STRING,
                                                         GDK_TYPE_PIXBUF,
                                                         G_TYPE_UINT,
                                                         G_TYPE_FLOAT,
                                                         G_TYPE_INT,
                                                         G_TYPE_UINT,
                                                         G_TYPE_UINT,
                                                         G_TYPE_STRING,
                                                         G_TYPE_UINT));

    project->fd = fd;
    project->name = strdup(name);
    project->id = id;

    project->gconfclient = client;

	return project;
}

void project_delete(Project *project)
{
    g_object_unref (G_OBJECT (project->gconfclient));
	free(project->name);
	free(project);
}

GtkWidget *project_view_new (Project *project)
{
	GtkWidget *vbox, *hbox, *button_hbox; 
	GtkWidget *sw;
	GdkPixbuf *pixbuf;
	gint w, h;
//	GtkObject *adjustment;

	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw),
										GTK_SHADOW_IN);

	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
								   GTK_POLICY_AUTOMATIC,
								   GTK_POLICY_AUTOMATIC);

	// create tree view
	project->tree_view = gtk_tree_view_new_with_model(project->model);
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(project->tree_view), TRUE);
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(
					GTK_TREE_VIEW(project->tree_view)), GTK_SELECTION_MULTIPLE);

	// remove the reference that the tree view takes
	g_object_unref(G_OBJECT(project->model));

	gtk_container_add(GTK_CONTAINER(sw), project->tree_view);

	// add columns to the tree view
	project_view_add_columns (project);

    /* get the selection */
    project->selection = gtk_tree_view_get_selection (
                                        GTK_TREE_VIEW (project->tree_view));

	g_signal_connect (G_OBJECT (project->selection), "changed",
                      GTK_SIGNAL_FUNC (project_selection_changed), project);

	if (project->pixbuf_exec == 0)
		project->pixbuf_exec = gtk_widget_render_icon(project->tree_view,
									GTK_STOCK_EXECUTE, GTK_ICON_SIZE_MENU, NULL);
	if (project->pixbuf_stop == 0)
		project->pixbuf_stop = gtk_widget_render_icon(project->tree_view,
									GTK_STOCK_STOP, GTK_ICON_SIZE_MENU, NULL);

	gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, &w, &h);

	if (project->pixbuf_queue == 0)
	{
		pixbuf = gtk_widget_render_icon(project->tree_view,
									GTK_STOCK_DND_MULTIPLE, GTK_ICON_SIZE_DND, NULL);

		project->pixbuf_queue = gdk_pixbuf_scale_simple(pixbuf, w, h, GDK_INTERP_BILINEAR);
		g_object_unref(pixbuf);
	}

	if (project->pixbuf_warning == 0)
	{
		pixbuf = gtk_widget_render_icon(project->tree_view,
									GTK_STOCK_DIALOG_WARNING, GTK_ICON_SIZE_DIALOG, NULL);

		project->pixbuf_warning = gdk_pixbuf_scale_simple(pixbuf, w, h, GDK_INTERP_BILINEAR);
		g_object_unref(pixbuf);
	}

	if (project->pixbuf_error == 0)
	{
		pixbuf = gtk_widget_render_icon(project->tree_view,
									GTK_STOCK_DIALOG_ERROR, GTK_ICON_SIZE_DIALOG, NULL);

		project->pixbuf_error = gdk_pixbuf_scale_simple(pixbuf, w, h, GDK_INTERP_BILINEAR);
		g_object_unref(pixbuf);
	}

	if (project->pixbuf_done == 0)
	{
		pixbuf = gtk_widget_render_icon(project->tree_view,
									GTK_STOCK_YES, GTK_ICON_SIZE_BUTTON, NULL);

		project->pixbuf_done = gdk_pixbuf_scale_simple(pixbuf, w, h, GDK_INTERP_BILINEAR);
		g_object_unref(pixbuf);
	}

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), sw, TRUE, TRUE, 0);

	hbox = gtk_hbox_new(FALSE, 5);

	project->auto_button = gtk_button_new();
	gtk_button_set_relief(GTK_BUTTON(project->auto_button), GTK_RELIEF_NONE);
	g_signal_connect(G_OBJECT(project->auto_button), "clicked",
					GTK_SIGNAL_FUNC(project_download_auto_send), project);
	button_hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(project->auto_button), button_hbox);

	if (project->auto_yes == 0)
		project->auto_yes = gtk_image_new_from_stock(GTK_STOCK_GO_DOWN,
													GTK_ICON_SIZE_MENU);
//FIXME: remove
	gtk_box_pack_start(GTK_BOX(button_hbox), project->auto_yes, FALSE, FALSE, 0);

	if (project->auto_no == 0)
		project->auto_no = gtk_image_new_from_stock(GTK_STOCK_CANCEL,
													GTK_ICON_SIZE_MENU);
//FIXME: remove
	gtk_box_pack_start(GTK_BOX(button_hbox), project->auto_no, FALSE, FALSE, 0);
	
	project->auto_label = gtk_label_new(NULL);
	gtk_box_pack_start(GTK_BOX(button_hbox), project->auto_label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), project->auto_button, FALSE, FALSE, 0);

//	project->status = gtk_label_new("Simultaneous downloads:");
//	gtk_box_pack_start(GTK_BOX(hbox), project->status, FALSE, FALSE, 0);
//	g_object_set (project->status, "xalign", 0.0, NULL);

//	adjustment = gtk_adjustment_new(project->simultaneous, 0, 20, 1, 5, 1);
//	project->spin = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 0.1, 0);
//	gtk_box_pack_start(GTK_BOX(hbox), project->spin, FALSE, FALSE, 0);

	project->status = gtk_label_new(NULL);
	gtk_box_pack_start(GTK_BOX(hbox), project->status, TRUE, TRUE, 0);
	project->status = gtk_label_new(NULL);
	gtk_box_pack_start(GTK_BOX(hbox), project->status, FALSE, FALSE, 5);

	g_signal_connect(G_OBJECT(project->status), "size_request",
					GTK_SIGNAL_FUNC(status_label_size_request), NULL);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	gtk_widget_show_all(vbox);

	return vbox;
}

void status_label_size_request(GtkWidget *widget, GtkRequisition  *requisition, gpointer data)
{
	if (requisition->width < widget->allocation.width)
		requisition->width = widget->allocation.width;
}

void
project_selection_changed (GtkTreeSelection *selection, gpointer data)
{
    Project *project = PROJECT (data);
    GList * selected;
    GtkTreePath *path;
    ProjectItem *item = g_malloc (sizeof (ProjectItem));
    GtkTreeIter iter;
    gint selected_rows;

    if (!project->download_properties)
        return;

    selected_rows = gtk_tree_selection_count_selected_rows (selection);

    switch (selected_rows)
    {
        case 0:
            download_properties_unset (project->download_properties, selected_rows);
            break;

        case 1:
            selected = gtk_tree_selection_get_selected_rows (selection, NULL);
            path = selected->data;
            if (gtk_tree_model_get_iter (project->model, &iter, path) == 0)
            {
                g_printerr ("No iter for path in item_update!\n");
                gtk_tree_path_free(path);
                break;
            } 

            gtk_tree_model_get (project->model, &iter,
                                COLUMN_ID, &item->id,
                                COLUMN_URL, &item->url,
                                COLUMN_SPEEDLIMIT, &item->speedlimit,
                                -1);

            item->project_id = project->id;

            download_properties_set (project->download_properties,
                                     item);
            break;

        default:
            download_properties_unset (project->download_properties, selected_rows);
            break;
    }
}

void
column_width_changed (GtkTreeViewColumn *column, GParamSpec *pspec, gpointer data)
{
    GConfClient *client;
    gint width;
    gchar *config_key;

    width = gtk_tree_view_column_get_width (column);

    client = g_object_get_data (G_OBJECT (column), "client");
    config_key = g_object_get_data (G_OBJECT (column), "config_key");
    g_object_set_data (G_OBJECT (column), "resizing", GINT_TO_POINTER (1));
    gconf_client_set_int (client, config_key, width, NULL);
}

static void
column_visible_config_notify (GConfClient *client,
                              guint        cnxn_id,
                              GConfEntry  *entry,
                              gpointer     user_data)
{
    GtkTreeViewColumn *column = user_data;

    g_return_if_fail (GTK_IS_TREE_VIEW_COLUMN (column));

    if (gconf_entry_get_value (entry) == NULL)
    {
        return;
    }
    else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_BOOL)
    {
        gtk_tree_view_column_set_visible (column,
                        gconf_value_get_bool (gconf_entry_get_value (entry)));
    }
}

static void
column_size_config_notify (GConfClient *client,
                           guint        cnxn_id,
                           GConfEntry  *entry,
                           gpointer     user_data)
{
    GtkTreeViewColumn *column = user_data;

    g_return_if_fail (GTK_IS_TREE_VIEW_COLUMN (column));

    if (GPOINTER_TO_INT (g_object_get_data (G_OBJECT (column), "resizing")))
    {
        g_object_set_data (G_OBJECT (column), "resizing", GINT_TO_POINTER (0));
        return;
    }

    if (gconf_entry_get_value (entry) == NULL)
    {
        return;
    }
    else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_INT)
    {
        gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_FIXED);
        gtk_tree_view_column_set_fixed_width (column,
                        gconf_value_get_int (gconf_entry_get_value (entry)));
    }
}

static void
column_widget_destroy_callback (GtkWidget *widget, gpointer data)
{
    guint notify_id;
    GConfClient *client;

    client = g_object_get_data (G_OBJECT (widget), "client");

    notify_id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (widget),
                                  "visible_notify_id"));
    if (notify_id != 0)
        gconf_client_notify_remove (client, notify_id);

    notify_id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (widget),
                                  "size_notify_id"));
    if (notify_id != 0)
        gconf_client_notify_remove (client, notify_id);
}

void
column_configure_visibility (GtkTreeViewColumn *column,
                             GConfClient       *client,
                             gchar             *config_key,
                             gboolean           visible_default)
{
    gboolean visible;
    guint notify_id;
    GConfValue* value = NULL;

    value = gconf_client_get (client, config_key, NULL);
    if (value)
    {
        visible = gconf_client_get_bool (client,
                                         config_key,
                                         NULL);
    }
    else
    {
        visible = visible_default;
    }

    gtk_tree_view_column_set_visible (column, visible);

    notify_id = gconf_client_notify_add (client,
                                         config_key,
                                         column_visible_config_notify,
                                         column,
                                         NULL, NULL);
    g_object_set_data (G_OBJECT (column), "visible_notify_id",
                       GUINT_TO_POINTER (notify_id));
}

void
column_configure_size (GtkTreeViewColumn *column,
                       GConfClient       *client,
                       gchar             *config_key,
                       gint               default_size)
{
    guint notify_id;
    gint size;

    size = gconf_client_get_int (client,
                                 config_key,
                                 NULL);
    if (!size)
        size = default_size;

    notify_id = gconf_client_notify_add (client,
                                         config_key,
                                         column_size_config_notify,
                                         column,
                                         NULL, NULL);
    g_object_set_data (G_OBJECT (column), "size_notify_id",
                       GUINT_TO_POINTER (notify_id));

    g_object_set (column, "resizable", TRUE, NULL);
    gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_FIXED);
    gtk_tree_view_column_set_fixed_width (column, size);

    g_object_set_data (G_OBJECT (column), "client", client);
    g_object_set_data (G_OBJECT (column), "config_key", config_key);

    g_signal_connect (G_OBJECT (column), "destroy",
                      G_CALLBACK (column_widget_destroy_callback),
                      NULL);

    g_signal_connect (column, "notify::width",
                      G_CALLBACK (column_width_changed), NULL);
}

static void
project_view_add_columns (Project *project)
{
    GtkTreeView *treeview = GTK_TREE_VIEW (project->tree_view);
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
    GConfClient *client = project->gconfclient;
    gchar *config_key;

    /* mime-type icon and file name */
    column = gtk_tree_view_column_new ();
    gtk_tree_view_column_set_title (column, "File name");

    renderer = gtk_cell_renderer_pixbuf_new ();
    gtk_tree_view_column_pack_start (column, renderer, FALSE);
    gtk_tree_view_column_set_attributes (column, renderer, "pixbuf",
                                         COLUMN_MIMETYPE_ICON,
                                         NULL);

    renderer = gtk_cell_renderer_text_new ();
    gtk_tree_view_column_pack_start (column, renderer, TRUE);
    gtk_tree_view_column_set_attributes (column, renderer, "text",
                                         COLUMN_FILE,
                                         NULL);

    /* filename visibility */
    config_key = "/apps/gdownman/columns/filename_visible";
    column_configure_visibility (column, client, config_key, TRUE);

    /* filename size */
    config_key = "/apps/gdownman/columns/filename_size";
    column_configure_size (column, client, config_key, 200);
    gtk_tree_view_append_column (treeview, column);

    /* status icon and status string */
    column = gtk_tree_view_column_new ();
    gtk_tree_view_column_set_title (column, "Status");

    renderer = gtk_cell_renderer_pixbuf_new ();
    gtk_tree_view_column_pack_start (column, renderer, FALSE);
    gtk_tree_view_column_set_attributes (column, renderer, "pixbuf",
                                         COLUMN_STATUS_ICON,
                                         NULL);

    renderer = gtk_cell_renderer_text_new ();
    gtk_tree_view_column_pack_start (column, renderer, TRUE);
    gtk_tree_view_column_set_attributes (column, renderer, "text",
                                         COLUMN_STATUS,
                                         NULL);
    gtk_tree_view_column_set_cell_data_func (column, renderer,
                            (GtkTreeCellDataFunc) project_column_format_status,
                            GINT_TO_POINTER (COLUMN_STATUS),
                            NULL);

    /* status visibility */
    config_key = "/apps/gdownman/columns/status_visible";
    column_configure_visibility (column, client, config_key, TRUE);

    /* status size */
    config_key = "/apps/gdownman/columns/status_size";
    column_configure_size (column, client, config_key, 100);
    gtk_tree_view_append_column (treeview, column);

    /* progressbar */
    renderer = gtk_cell_renderer_progress_new ();
    g_object_set (renderer, "xalign", 1.0, NULL);
    column = gtk_tree_view_column_new_with_attributes ("Progress", renderer,
                                                       "value",
                                                       COLUMN_PERCENT,
                                                       NULL);
    gtk_tree_view_column_set_cell_data_func (column, renderer,
                        (GtkTreeCellDataFunc) project_column_format_progress,
                        GINT_TO_POINTER (COLUMN_PERCENT),
                        NULL);

    /* progressbar visibility */
    config_key = "/apps/gdownman/columns/progressbar_visible";
    column_configure_visibility (column, client, config_key, TRUE);

    /* progressbar size */
    config_key = "/apps/gdownman/columns/progressbar_size";
    column_configure_size (column, client, config_key, 150);
    gtk_tree_view_append_column (treeview, column);

    /* speed */
    renderer = gtk_cell_renderer_text_new ();
    g_object_set (renderer, "xalign", 1.0, NULL);
    column = gtk_tree_view_column_new_with_attributes ("Speed",
                                                       renderer,
                                                       "text",
                                                       COLUMN_SPEED,
                                                       NULL);
    gtk_tree_view_column_set_cell_data_func (column, renderer,
                            (GtkTreeCellDataFunc) project_column_format_speed,
                            GINT_TO_POINTER (COLUMN_SPEED),
                            NULL);

    /* speed visibility */
    config_key = "/apps/gdownman/columns/speed_visible";
    column_configure_visibility (column, client, config_key, TRUE);

    /* speed size */
    config_key = "/apps/gdownman/columns/speed_size";
    column_configure_size (column, client, config_key, 50);
    gtk_tree_view_append_column (treeview, column);

    /* downloaded */
    renderer = gtk_cell_renderer_text_new ();
    g_object_set (renderer, "xalign", 1.0, NULL);
    column = gtk_tree_view_column_new_with_attributes ("Downloaded",
                                                       renderer,
                                                       "text",
                                                       COLUMN_DOWNLOADED,
                                                       NULL);
    gtk_tree_view_column_set_cell_data_func (column, renderer,
                            (GtkTreeCellDataFunc) project_column_format_value,
                            GINT_TO_POINTER (COLUMN_DOWNLOADED),
                            NULL);

    /* downloaded visibility */
    config_key = "/apps/gdownman/columns/downloaded_visible";
    column_configure_visibility (column, client, config_key, FALSE);

    /* downloaded size */
    config_key = "/apps/gdownman/columns/downloaded_size";
    column_configure_size (column, client, config_key, 50);
    gtk_tree_view_append_column (treeview, column);

    /* total size */
    renderer = gtk_cell_renderer_text_new ();
    g_object_set (renderer, "xalign", 1.0, NULL);
    column = gtk_tree_view_column_new_with_attributes ("File size",
                                                       renderer,
                                                       "text",
                                                       COLUMN_TOTAL,
                                                       NULL);
    gtk_tree_view_column_set_cell_data_func (column, renderer,
                            (GtkTreeCellDataFunc) project_column_format_value,
                            GINT_TO_POINTER (COLUMN_TOTAL), NULL);

    /* total size visibility */
    config_key = "/apps/gdownman/columns/total_size_visible";
    column_configure_visibility (column, client, config_key, TRUE);

    /* total size size */
    config_key = "/apps/gdownman/columns/total_size_size";
    column_configure_size (column, client, config_key, 50);
    gtk_tree_view_append_column (treeview, column);

    /* url */
    renderer = gtk_cell_renderer_text_new ();
    column = gtk_tree_view_column_new_with_attributes ("URL",
                                                       renderer,
                                                       "text",
                                                       COLUMN_URL,
                                                       NULL);
    /* url visibility */
    config_key = "/apps/gdownman/columns/url_visible";
    column_configure_visibility (column, client, config_key, TRUE);

    /* url size */
    config_key = "/apps/gdownman/columns/url_size";
    column_configure_size (column, client, config_key, 200);
    gtk_tree_view_append_column (treeview, column);
}

void project_column_format_value(GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
				GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data)
{
	gint column = GPOINTER_TO_INT(data);
	GValue value = {0, };
	char buf[10];
	guint size;

	gtk_tree_model_get_value (tree_model, iter, column, &value);

	switch (G_TYPE_FUNDAMENTAL(G_VALUE_TYPE(&value)))
	{
		case G_TYPE_FLOAT:
			snprintf(buf, 10, "%.2f", g_value_get_float(&value));
			break;

		case G_TYPE_UINT:
			size = g_value_get_uint (&value);

			if (size >= 1024*1024)
			{
				snprintf(buf, 10, "%.1f Mb", ((float)size/(1024*1024)));
			}
			else if (size >= 1024)
			{
				snprintf(buf, 10, "%.1f Kb", ((float)size/1024));
			}
			else
				snprintf(buf, 10, "%i b", size);
			break;

		default:
			snprintf(buf, 10, "error");
	}

	g_object_set (cell, "text", buf, NULL);

	g_value_unset (&value);
}

void
project_column_format_progress (GtkTreeViewColumn *tree_column,
    GtkCellRenderer *cell,
    GtkTreeModel *tree_model,
    GtkTreeIter *iter,
    gpointer data)
{
    gint column = GPOINTER_TO_INT(data);
    GValue value = {0, };

    gtk_tree_model_get_value (tree_model, iter, column, &value);

    g_object_set (cell, "value", g_value_get_float(&value) / 100, NULL);

    g_value_unset (&value);
}

void project_column_format_status(GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
				GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data)
{
	gint column = GPOINTER_TO_INT(data);
	GValue value = {0, };
	guint status;

	gtk_tree_model_get_value (tree_model, iter, column, &value);

	status = g_value_get_uint (&value);

//FIXME: Flag to know if status changed to avoid the text update
	g_object_set (cell, "text", dm_transfer_status_string(status), NULL);

	g_value_unset (&value);
}

void project_column_format_speed(GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
				GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data)
{
	gint column = GPOINTER_TO_INT(data);
	GValue value = {0, };
	char buf[10];
	guint speed;

	gtk_tree_model_get_value (tree_model, iter, column, &value);

	speed = g_value_get_int (&value);

	if (speed == -1)
		buf[0] = '\0';
//	else if (speed > 1024*1024)
//	{
//		snprintf(buf, 10, "%.1f Mb/s", ((float)speed/(1024*1024)));
//	}
	else if (speed >= 1024)
	{
		snprintf(buf, 10, "%.1f Kb/s", ((float)speed/1024));
	}
	else
		snprintf(buf, 10, "%i b/s", speed);

	g_object_set (cell, "text", buf, NULL);

	g_value_unset (&value);
}

void project_download_foreach(GtkTreeModel *model, GtkTreePath *path,
								GtkTreeIter *iter, gpointer data)
{
	Project *project = (Project *)data;
	int id;

	gtk_tree_model_get(model, iter, COLUMN_ID, &id, -1);

	switch(project->action)
	{
		case DOWNLOAD_START:
//FIXME: don't sent if already started
            dm_command_download_start (project->fd, project->id, id);
			break;

		case DOWNLOAD_STOP:
//FIXME: don't sent if already stopped
            dm_command_download_stop (project->fd, project->id, id);
			break;

		case DOWNLOAD_QUEUE:
//FIXME: don't sent if already queued
            dm_command_download_queue (project->fd, project->id, id);
			break;

		case DOWNLOAD_DELETE:
            dm_command_download_delete (project->fd, project->id, id);
			break;

		default:
			break;
	}
}

void project_download(Project *project, int action, int fd)
{
	GtkTreeSelection *selection;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(project->tree_view));

	project->action = action;
	project->fd = fd;

	gtk_tree_selection_selected_foreach(selection,
		(GtkTreeSelectionForeachFunc) project_download_foreach, project);
}

void project_download_auto_send(GtkWidget *caller, gpointer data)
{
    Project *project = (Project *) data;

    dm_command_auto (project->fd, project->id,
                     !project_settings_autodownload_get (project->settings));
}

void project_download_auto_set(Project *project)
{
    if (project_settings_autodownload_get (project->settings))
	{
		gtk_widget_hide(project->auto_no);
		gtk_widget_show(project->auto_yes);
//		gtk_label_set_text(GTK_LABEL(project->auto_label),
//					"Queued items are downloaded automatically");
	}
	else
	{
		gtk_widget_hide(project->auto_yes);
		gtk_widget_show(project->auto_no);
//		gtk_label_set_text(GTK_LABEL(project->auto_label),
//					"Queued items must be started manually");
	}
}

void
project_maxdownloads_send (Project *project, guint maxdownloads)
{
    project_settings_maxdownloads_set (project->settings, maxdownloads);

    dm_command_maxdownloads (project->fd, project->id, maxdownloads);
}

void project_select_all(Project *project)
{
	GtkTreeSelection *selection;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(project->tree_view));
	gtk_tree_selection_select_all(selection);
}

void project_select_status(Project *project, gint status)
{
    GtkTreeSelection *selection;
    GtkTreeIter iter;
	gint download_status;

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(project->tree_view));

    if (gtk_tree_model_get_iter_first (project->model, &iter))
    {
        do {
            gtk_tree_model_get (project->model, &iter,
                                COLUMN_STATUS, &download_status, -1);
            if (download_status == status)
                gtk_tree_selection_select_iter(selection, &iter);
            else
                gtk_tree_selection_unselect_iter(selection, &iter);

        } while (gtk_tree_model_iter_next (project->model, &iter));
    }
}

void
project_select_active (Project *project)
{
    GtkTreeSelection *selection;
    GtkTreeIter iter;
	gint download_status;

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(project->tree_view));

    if (gtk_tree_model_get_iter_first (project->model, &iter))
    {
        do {
            gtk_tree_model_get (project->model, &iter,
                                COLUMN_STATUS, &download_status, -1);
            if ((download_status >= STATUS_INIT) &&
                (download_status <= STATUS_TRANSFERING))
                gtk_tree_selection_select_iter(selection, &iter);
            else
                gtk_tree_selection_unselect_iter(selection, &iter);

        } while (gtk_tree_model_iter_next (project->model, &iter));
    }
}

void
project_select_matched  (Project *project, const gchar *pattern)
{
    GtkTreeSelection *selection;
    GtkTreeIter iter;
    GPatternSpec *pspec;
	gchar *filename;

    pspec = g_pattern_spec_new (pattern);

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(project->tree_view));

    if (gtk_tree_model_get_iter_first (project->model, &iter))
    {
        do {
            gtk_tree_model_get (project->model, &iter,
                                COLUMN_FILE, &filename, -1);
            if (g_pattern_match_string (pspec, filename))
                gtk_tree_selection_select_iter(selection, &iter);
            else
                gtk_tree_selection_unselect_iter(selection, &iter);

        } while (gtk_tree_model_iter_next (project->model, &iter));
    }
}

void
project_queue_item (Project *project, ProjectItem *item)
{
    if (!project->updates_waiting)
    {
        /* Wait a bit to chunk updates to the view */
        g_timeout_add (100, project_process_updates, project);
    }
    project->updates_waiting += 1;

    g_queue_push_tail (project->updates_queue, item);

    if (project->updates_waiting > 199)
        project_process_updates (project);
}

gboolean
project_process_updates (gpointer data)
{
    Project *project = PROJECT (data);
    ProjectItem *item;
    gint i = 0;

    while (!g_queue_is_empty (project->updates_queue))
    {
        item = g_queue_pop_head (project->updates_queue);
        switch (item->type)
        {
            case PROJECT_ITEM_NEW:
                project_item_new (project, item);
                g_free (item->url);
                g_free (item);
                break;

            case PROJECT_ITEM_UPDATE:
            case PROJECT_ITEM_DONE:
            case PROJECT_ITEM_STOP:
                project_item_update (project, item);
                g_free (item);
                break;

            case PROJECT_ITEM_DELETE:
                project_item_delete (project, item->id);
                break;

            case PROJECT_ITEM_SPEEDLIMIT:
                project_item_speedlimit (project, item);
                break;

            default:
                g_printerr ("Unknow update type\n");
        }
        i++;
    }

    project->updates_waiting = 0;
    return FALSE;
}

void
project_item_new (Project *project, ProjectItem *item)
{
	GtkTreeIter iter;
	GtkTreePath *path;
	GtkTreeRowReference *row;
	
	gint w, h;
	GdkPixbuf *status_pixbuf;
    gchar *file;

	if (item->status == STATUS_QUEUED)
		status_pixbuf = project->pixbuf_queue;
	else if (item->status == STATUS_STOPPED)
		status_pixbuf = project->pixbuf_stop;
	else if (item->status == STATUS_DONE)
		status_pixbuf = project->pixbuf_done;
	else if ((item->status > ERROR_FIRST) && (item->status < ERROR_LAST))
		status_pixbuf = project->pixbuf_warning;
	else
		status_pixbuf = project->pixbuf_exec;

	gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, &w, &h);

    if (item->url[strlen (item->url) - 1] == '/')
        file = "index.html";
    else
        file = (strrchr(item->url, '/')) ? strrchr(item->url, '/') + 1 : item->url;

    gtk_list_store_append(GTK_LIST_STORE(project->model), &iter);
    gtk_list_store_set(GTK_LIST_STORE(project->model), &iter,
                    COLUMN_ID, item->id,
					COLUMN_MIMETYPE_ICON, item->mimetype_pixbuf,
                    COLUMN_FILE, file,
                    COLUMN_FILENAME, item->localpath,
                    COLUMN_STATUS_ICON, status_pixbuf,
                    COLUMN_STATUS, item->status,
                    COLUMN_PERCENT, item->percent,
                    COLUMN_SPEED, -1,
                    COLUMN_DOWNLOADED, (long int)item->downloaded,
                    COLUMN_TOTAL, (long int)item->total,
                    COLUMN_URL, item->url,
                    COLUMN_SPEEDLIMIT, item->speedlimit,
                    -1);

	path = gtk_tree_model_get_path(project->model, &iter);
	row = gtk_tree_row_reference_new(project->model, path);
    g_hash_table_insert (project->hash, GINT_TO_POINTER (item->id), row);
	gtk_tree_path_free(path);
}

void
project_item_update (Project *project, ProjectItem *item)
{
	GtkTreeIter iter;
	GtkTreePath *path;
	GtkTreeRowReference *row;
	int status2;
	char *file;

    row = g_hash_table_lookup (project->hash, GINT_TO_POINTER (item->id));

	if (row)
	{
		path = gtk_tree_row_reference_get_path(row);
		if (gtk_tree_model_get_iter(project->model, &iter, path) == 0)
		{
			printf("No iter for path in item_update!\n");
			gtk_tree_path_free(path);
			return;
		}
		gtk_tree_path_free(path);

        gtk_tree_model_get(project->model, &iter,
                           COLUMN_FILE, &file,
                           COLUMN_STATUS, &status2, -1);

        if (item->status == status2)
        {
            gtk_list_store_set(GTK_LIST_STORE(project->model), &iter,
                               COLUMN_PERCENT, item->percent,
                               COLUMN_SPEED, item->speed,
                               COLUMN_DOWNLOADED, (long int)item->downloaded,
                               COLUMN_TOTAL, (long int)item->total,
                               -1);
        }
        else
        {
            GdkPixbuf *status_pixbuf;
            gint done = 0;

			if (item->status == STATUS_QUEUED)
				status_pixbuf = project->pixbuf_queue;
			else if (item->status == STATUS_STOPPED)
				status_pixbuf = project->pixbuf_stop;
			else if (item->status == STATUS_DONE)
			{
                status_pixbuf = project->pixbuf_done;
				done = 1;
			}
			else if ((item->status > ERROR_FIRST) && (item->status < ERROR_LAST))
				status_pixbuf = project->pixbuf_warning;
			else
				status_pixbuf = project->pixbuf_exec;

            if (done)
                gtk_list_store_set(GTK_LIST_STORE(project->model), &iter,
//						COLUMN_MIMETYPE_ICON, item->mimetype_pixbuf,
                        COLUMN_STATUS_ICON, status_pixbuf,
                        COLUMN_STATUS, item->status,
                        COLUMN_PERCENT, item->percent,
                        COLUMN_SPEED, item->speed,
                        COLUMN_DOWNLOADED, (long int)item->downloaded,
                        COLUMN_TOTAL, (long int)item->total,
                        -1);
            else
                gtk_list_store_set(GTK_LIST_STORE(project->model), &iter,
                        COLUMN_STATUS_ICON, status_pixbuf,
                        COLUMN_STATUS, item->status,
                        COLUMN_PERCENT, item->percent,
                        COLUMN_SPEED, item->speed,
                        COLUMN_DOWNLOADED, (long int)item->downloaded,
                        COLUMN_TOTAL, (long int)item->total,
                        -1);
        }
        g_free(file);
    }
    else
    {
        g_printerr ("ERROR: Item id = %i not found for update!!!\n", item->id);
    }
}

void project_item_delete(Project *project, int id)
{
	GtkTreeIter iter;
	GtkTreePath *path;
	GtkTreeRowReference *row;

    row = g_hash_table_lookup (project->hash, GINT_TO_POINTER (id));

	if (row)
	{
		path = gtk_tree_row_reference_get_path(row);
		if (gtk_tree_model_get_iter(project->model, &iter, path) == 0)
		{
			printf("No iter for path in item_delete!\n");
			return;
		}

		gtk_list_store_remove(GTK_LIST_STORE(project->model), &iter);
		gtk_tree_path_free(path);
		gtk_tree_row_reference_free(row);
        
		g_hash_table_remove (project->hash, GINT_TO_POINTER (id));
	}
	else
	{
		printf("ERROR: Item not found for delete!!!\n");
	}
}

void
project_item_speedlimit (Project *project, ProjectItem *item)
{
	GtkTreeIter iter;
	GtkTreePath *path;
	GtkTreeRowReference *row;

    row = g_hash_table_lookup (project->hash, GINT_TO_POINTER (item->id));

	if (row)
	{
		path = gtk_tree_row_reference_get_path (row);
		if (gtk_tree_model_get_iter (project->model, &iter, path) == 0)
		{
			printf ("No iter for path in item_update!\n");
			gtk_tree_path_free (path);
			return;
		}
		gtk_tree_path_free (path);

        gtk_list_store_set (GTK_LIST_STORE (project->model), &iter,
                            COLUMN_SPEEDLIMIT, item->speedlimit,
                            -1);

//FIXME: need to update dialogs? speedlimit_changed signal?
//       think of "confirmation pending" so GUI is more responsive
    }
    else
    {
        g_printerr ("ERROR: Item id = %i not found for update!!!\n", item->id);
    }
}

void
project_path_set (Project *project, gchar *path)
{
    project_settings_path_set (project->settings, path);

//FIXME: move to signals in ProjectSettings
//    g_signal_emit (G_OBJECT (project),
//                   project_signals[PROJECT_PATH],
//                   0);
}

void
project_auto_set (Project *project, guint auto_download)
{
    project_settings_autodownload_set (project->settings, auto_download);

    /* Update the icon */
    project_download_auto_set(project);

    g_signal_emit (G_OBJECT (project),
                   project_signals[PROJECT_AUTODOWNLOAD],
                   0);
}

void
project_maxdownloads_set (Project *project, guint maxdownloads)
{
    project_settings_maxdownloads_set (project->settings, maxdownloads);

    /* Update the spin */
//FIXME:
//    project_maxdownloads_set(project);

    g_signal_emit (G_OBJECT (project),
                   project_signals[PROJECT_MAXDOWNLOADS],
                   0);
}

void project_status_set(Project *project, char *status)
{
	gtk_label_set_text(GTK_LABEL(project->status), status);
}

void
project_download_properties (Project *project,
    DownloadProperties *download_properties)
{
    project->download_properties = download_properties;
    project_selection_changed (project->selection, project);
}

const char *project_name(Project *project)
{
	return project->name;
}

int project_id(Project *project)
{
	return project->id;
}
