#ifndef __PROJECT_H__
#define __PROJECT_H__

#include <glib-object.h>
#include <gconf/gconf-client.h>

#include <project-item.h>
#include <project-settings.h>
#include <download-properties.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

typedef enum {
	DOWNLOAD_START = 0,
	DOWNLOAD_STOP,
	DOWNLOAD_QUEUE,
	DOWNLOAD_DELETE
} DownloadCommand;


#define PROJECT(obj)           G_TYPE_CHECK_INSTANCE_CAST (obj, project_get_type (), Project)
#define PROJECT_CLASS(klass)   G_TYPE_CHECK_CLASS_CAST (klass, project_get_type (), ProjectClass)
#define IS_PROJECT(obj)        G_TYPE_CHECK_INSTANCE_TYPE (obj, project_get_type ())


typedef struct _Project       Project;
typedef struct _ProjectClass  ProjectClass;

struct _Project
{
    GObject object;

    /*< private >*/

	GtkTreeModel *model;
	GtkWidget *tree_view;
    GtkTreeSelection *selection;
	GtkWidget *status;

	GtkWidget *auto_button;
	GtkWidget *auto_label;
	GtkWidget *auto_yes;
	GtkWidget *auto_no;

    guint updates_waiting;
    GQueue *updates_queue;

	GHashTable *hash;
	int action;
	int fd;

	GdkPixbuf *pixbuf_mimetype;
	GdkPixbuf *pixbuf_exec;
	GdkPixbuf *pixbuf_stop;
	GdkPixbuf *pixbuf_queue;
	GdkPixbuf *pixbuf_warning;
	GdkPixbuf *pixbuf_error;
	GdkPixbuf *pixbuf_done;

    gchar *name;
    guint id;

    /* settings */
    ProjectSettings *settings;

    /* configuration */
    GConfClient *gconfclient;

    /* dialogs */
    DownloadProperties *download_properties;
};

struct _ProjectClass
{
    GObjectClass parent_class;

    void (* project_autodownload_signal)     (void);
    void (* project_maxdownloads_signal)     (void);
};

GType           project_get_type        (void);
Project*        project_new             (guint id, gchar *name, guint fd,
                                         GConfClient *client);

void            project_delete          (Project *project);

GtkWidget *     project_view_new        (Project *project);
const gchar *   project_name            (Project *project);
gint            project_id              (Project *project);

void    project_download        (Project *project, int action, int fd);
void    project_select_all      (Project *project);
void    project_select_status   (Project *project, gint status);
void    project_select_active   (Project *project);
void    project_select_matched  (Project *project, const gchar *pattern);

void    project_queue_item      (Project *project, ProjectItem *item);
void    project_item_delete     (Project *project, int id);

void    project_status_set          (Project *project, char *status);
void    project_path_set            (Project *project, gchar *path);
void    project_auto_set            (Project *project, guint auto_download);
void    project_maxdownloads_set    (Project *project, guint maxdownloads);

void    project_download_auto_send  (GtkWidget *caller, gpointer data);
void    project_maxdownloads_send   (Project *project, guint maxdownloads);

void    project_download_properties (Project *project,
                                     DownloadProperties *download_properties);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __PROJECT_H__ */
