#include <glib.h>

#include <libxml/uri.h>
#include <libxml/HTMLtree.h>
#include <libxml/HTMLparser.h>

#include "parser.h"

void    extract_links   (xmlNodePtr node, GHashTable *hash, gchar *base_url);


GHashTable *
parse_html_file (GnomeVFSURI *target_uri, GnomeVFSURI *source_uri)
{
    GHashTable *hash;
    htmlDocPtr doc;
    gchar *filename;

    if (!gnome_vfs_uri_is_local (target_uri))
        return NULL;

    filename = gnome_vfs_uri_to_string (target_uri, GNOME_VFS_URI_HIDE_PASSWORD);
    doc = htmlParseFile (filename, NULL);

    if (!doc)
        return NULL;

//    hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
    hash = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);

    if (doc->children)
    {
        extract_links (doc->children, hash,
            gnome_vfs_uri_to_string (source_uri, GNOME_VFS_URI_HIDE_PASSWORD));
    }

    return hash;
}

void
extract_links (xmlNodePtr node, GHashTable *hash, gchar *base_url)
{
    xmlChar *url = NULL;
    xmlChar *full_url;

	for (; node != NULL; node = node->next)
	{
        if (!node->name)
            continue;

        if (g_ascii_strcasecmp (node->name, "a") == 0)
        {
            url = xmlGetProp (node, "href");
        }
        else if (g_ascii_strcasecmp (node->name, "img") == 0)
        {
            url = xmlGetProp (node, "src");
        }

        if (url != NULL)
        {
//FIXME: use gnome-vfs here
            full_url = xmlBuildURI (url, base_url);
            if (full_url)
                g_hash_table_insert (hash, full_url, 0);
//            g_free (url);
        }

        if (node->children)
            extract_links (node->children, hash, base_url);
	}
}

/*
void
move_matching_strings (GHashTable *source, GHashTable *target, gchar *pattern)
{
    GPatternSpec *pspec;

    pspec = g_pattern_spec_new (pattern);

    if (g_pattern_match_string (pspec, string))
        move

}
*/
