#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <glib/gstring.h>
#include <glib/gspawn.h>

#include "commands.h"
#include "sockets.h"

gint
dm_client_local_connect (gchar *name)
{
	int sock_fd;
	struct sockaddr_un addr;
	size_t size;
	int flags;
	GString *buffer = g_string_new(NULL);
	char answer[10];

	sock_fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (sock_fd < 0)
		return -1;

	g_string_printf(buffer, "/tmp/downmand.%s", getpwuid(geteuid())->pw_name);

	// configure the socket
	addr.sun_family = AF_FILE;
	strcpy(addr.sun_path, buffer->str);
	size = sizeof(addr.sun_family) + strlen(addr.sun_path) + 1;

	if (connect(sock_fd, (struct sockaddr *)&addr, size) < 0)
		return -1;

	g_string_printf(buffer, "%u %s\n", DM_COMMAND_CONFIG_NAME, name);

	if (write(sock_fd, buffer->str, buffer->len) < 0)
	{
		close(sock_fd);
		g_string_free(buffer, TRUE);
		return -1;
	}
	else
		g_string_free(buffer, TRUE);

	if (read(sock_fd, answer, 8) < 0)
	{
		close(sock_fd);
		return -1;
	}

	if (strncmp(answer, "NAME OK\n", 8) != 0)
	{
		close(sock_fd);
		return -1;
	}

	flags = fcntl(sock_fd, F_GETFL, 0);
	if (flags == -1)
	{
		close(sock_fd);
		return -1;
	}
	flags |= O_NONBLOCK;
	if (fcntl(sock_fd, F_SETFL, flags) < 0)
	{
		close(sock_fd);
		return -1;
	}

	return sock_fd;
}

gboolean
dm_client_daemon_start (GError **error)
{
    char **argv = calloc (1, sizeof (char *));

    argv[0] = BINDIR "/downmand";

    if (!g_spawn_sync (NULL, argv, NULL, 0, 0, NULL, NULL,
        NULL, NULL, error))
    {
        return FALSE;
    }

    return TRUE;
}
